/*
 * The MIT License
 *
 * Copyright 2016-2020 Charles University
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */
package cz.cuni.mff.java.cv09;

import java.util.Arrays;
import java.util.concurrent.ForkJoinPool;

import static java.util.concurrent.ForkJoinTask.invokeAll;

import java.util.concurrent.RecursiveTask;

public class FindMax {

    public static void main(String[] args) {
        int[] array = {1, 0, 2, 9, 3, 8, 4, 7, 5, 6, 10, -1, -3, -2};
        System.out.println("Max via threads: " + findMaxViaThreads(array));
        System.out.println("Max via ForkJoin: " + findMaxViaForkJoin(array));
        System.out.println("Max via streams: " + findMaxViaStreams(array));
    }

    public static int findMaxViaThreads(int[] array) {
        int nThreads = Runtime.getRuntime().availableProcessors();

        int x = array.length / nThreads;
        FinderThread[] threads;
        if (x > 0) {
            System.out.println("Running with " + nThreads + " threads");
            threads = new FinderThread[nThreads];
            for (int i = 0; i < nThreads; i++) {
                int high = ((i + 1) * x) - 1;
                if (i == (nThreads - 1)) {
                    high = array.length - 1;
                }
                threads[i] = new FinderThread(array, i * x, high);
                threads[i].start();
            }
        } else {
            threads = new FinderThread[1];
            threads[0] = new FinderThread(array, 0, array.length - 1);
            threads[0].start();
        }
        int result = Integer.MIN_VALUE;
        for (FinderThread t : threads) {
            while (t.isAlive()) {
                try {
                    t.join();
                } catch (InterruptedException ex) {
                }
            }
            if (result < t.getResult()) {
                result = t.getResult();
            }
        }
        return result;
    }

    private static class FinderThread extends Thread {

        private int[] array;
        private int min;
        private int max;
        private int result = Integer.MIN_VALUE;

        public FinderThread(int[] array, int min, int max) {
            this.array = array;
            this.min = min;
            this.max = max;
            System.out.println("Limits " + min + "---" + max);
        }

        @Override
        public void run() {
            for (int i = min; i <= max; i++) {
                if (array[i] > result) {
                    result = array[i];
                }
            }
        }

        public int getResult() {
            return result;
        }
    }

    public static int findMaxViaForkJoin(int[] array) {
        FinderTask ft = new FinderTask(array, 0, array.length);
        return ForkJoinPool.commonPool().invoke(ft);
    }

    private static final int TRESHOLD = 4;

    private static class FinderTask extends RecursiveTask<Integer> {

        private int[] array;
        private int min;
        private int max;

        public FinderTask(int[] array, int min, int max) {
            this.array = array;
            this.min = min;
            this.max = max;
            System.out.println("Task for " + min + "---" + max);
        }

        @Override
        protected Integer compute() {
            if (max - min < TRESHOLD) {
                return computeDirectly();
            } else {

                FinderTask left = new FinderTask(array, min, min + (max - min) / 2);
                FinderTask right = new FinderTask(array, min + (max - min) / 2, max);
                invokeAll(left, right);
                return Integer.max(left.join(), right.join());
            }
        }

        private Integer computeDirectly() {
            int result = Integer.MIN_VALUE;
            for (int i = min; i < max; i++) {
                if (array[i] > result) {
                    result = array[i];
                }
            }
            return result;
        }

    }

    public static int findMaxViaStreams(int[] array) {
        if (array.length == 0) {
            throw new IllegalArgumentException("Empty array");
        }
        return Arrays.stream(array).parallel().max().getAsInt();
    }
}
