/*
 * The MIT License
 *
 * Copyright 2016-2020 Charles University
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */
package cz.cuni.mff.java.cv09;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.IOException;
import java.io.InputStreamReader;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.nio.file.StandardOpenOption;
import java.util.ArrayList;
import java.util.List;
import java.util.ListIterator;
import java.util.stream.Collectors;

public class ToDo {

    private static class Item {

        int priority;
        String text;

        public Item(int priority, String text) {
            this.priority = priority;
            this.text = text;
        }

        public Item(String line) {
            String[] items = line.split(":");
            try {
                priority = Integer.parseInt(items[0]);
            } catch (NumberFormatException ex) {
                throw new IllegalArgumentException("Bad line", ex);
            }
            text = items[1];
        }

        @Override
        public String toString() {
            return priority + ":" + text;
        }
    }

    private static List<Item> todos;
    private static Path fname = Path.of(System.getProperty("user.home"), ".javatodo");

    private static void readTodos() {
        if (!Files.exists(fname)) {
            System.out.println("The file does not exists; probably first run of the program");
            todos = new ArrayList<>();
        } else {
            try {
                todos = Files.lines(fname).map(Item::new).collect(Collectors.toList());
            } catch (IOException ex) {
                System.out.println("The file cannot be read");
                System.exit(1);
            }
        }
    }

    private static void writeTodos() {
        try (BufferedWriter f = Files.newBufferedWriter(fname, StandardOpenOption.CREATE, StandardOpenOption.WRITE, StandardOpenOption.TRUNCATE_EXISTING)) {
            for (Item i : todos) {
                f.write(i.toString());
                f.write("\n");
            }
        } catch (IOException ex) {
            System.out.println("The file cannot be written");
            System.exit(1);
        }
    }

    private static void listTodos(boolean reverse) {
        if (!reverse) {
            todos.forEach(System.out::println);
        } else {
            ListIterator<Item> li = todos.listIterator(todos.size());
            while (li.hasPrevious()) {
                System.out.println(li.previous());
            }
        }
    }

    private static void addTodo(String[] args) {
        if (args.length < 3) {
            System.out.println("Not enough parameters");
            System.exit(1);
        }
        int prior;
        try {
            prior = Integer.parseInt(args[1]);
        } catch (NumberFormatException ex) {
            prior = 0;
        }
        StringBuilder sb = new StringBuilder();
        for (int i = 2; i < args.length; i++) {
            sb.append(args[i]).append(" ");
        }
        Item item = new Item(prior, sb.toString());
        boolean added = false;
        for (int i = 0; i < todos.size(); i++) {
            if (todos.get(i).priority < prior) {
                todos.add(i, item);
                added = true;
                break;
            }
        }
        if (!added) {
            todos.add(item);
        }
    }

    private static void removeTodo() {
        for (int i = 0; i < todos.size(); i++) {
            System.out.println(i + " " + todos.get(i).text);
        }
        System.out.println("Item to remove: ");
        try {
            boolean repeat = true;
            do {
                BufferedReader input = new BufferedReader(new InputStreamReader(System.in));
                String line = input.readLine().trim();
                try {
                    int answer = Integer.parseInt(line);
                    if (answer >= 0 && answer < todos.size()) {
                        System.out.println("Removing item " + answer);
                        todos.remove(answer);
                        repeat = false;
                    } else {
                        System.out.println("Bad number, try again");
                    }
                } catch (NumberFormatException ex) {
                    System.out.println("Not a number, try again");
                }
            } while (repeat);
        } catch (IOException ex) {
            System.out.println("Cannot read from the user");
            System.exit(1);
        }
    }

    public static void main(String[] args) {
        if (args.length == 0) {
            System.out.println("No action specified");
            System.exit(1);
        }
        readTodos();
        switch (args[0]) {
            case ("-a") -> {
                addTodo(args);
                writeTodos();
            }
            case ("-l") -> listTodos(false);
            case ("-r") -> listTodos(true);
            case ("-d") -> {
                removeTodo();
                writeTodos();
            }
            default -> {
                System.out.println("Unknown action");
                System.exit(1);
            }
        }
    }

}
